from requests_mock import Mocker
from trailblazer.clients.slurm_api_client.dto import SlurmJobsResponse
from trailblazer.clients.slurm_api_client.slurm_api_client import SlurmApiClient


def test_get_slurm_jobs(client: SlurmApiClient, mock_request: Mocker, jobs_response: dict):
    # GIVEN a mocked jobs response
    mock_request.get(f"{client.base_url}/slurm/v0.0.40/jobs", json=jobs_response)

    # WHEN retrieving the jobs
    jobs_response: SlurmJobsResponse = client.get_jobs()

    # THEN the jobs are deserialized without error
    assert jobs_response



import base64
import os
from cryptography.hazmat.primitives.ciphers.aead import AESGCM

NONCE_SIZE = 12


def encrypt_data(data: str, secret_key: str) -> str:
    data: bytes = data.encode()
    cipher_text: bytes = _encrypt_with_aes(data=data, secret_key=secret_key)
    return _bytes_to_string(cipher_text)


def decrypt_data(data: str, secret_key: str) -> str:
    data: bytes = base64.b64decode(data)
    plain_text: bytes = _decrypt_with_aes(data=data, secret_key=secret_key)
    return plain_text.decode()


def _encrypt_with_aes(data: bytes, secret_key: str) -> bytes:
    cipher = AESGCM(base64.b64decode(secret_key))
    nonce: bytes = os.urandom(NONCE_SIZE)
    cipher_text: bytes = cipher.encrypt(nonce=nonce, data=data, associated_data=None)
    return nonce + cipher_text


def _decrypt_with_aes(data: bytes, secret_key: str) -> bytes:
    cipher = AESGCM(base64.b64decode(secret_key))
    nonce: bytes = data[:NONCE_SIZE]
    cipher_text: bytes = data[NONCE_SIZE:]
    return cipher.decrypt(nonce=nonce, data=cipher_text, associated_data=None)


def _bytes_to_string(data: bytes) -> str:
    return base64.b64encode(data).decode()



from datetime import datetime

from sqlalchemy.orm import Session

from trailblazer.constants import SlurmJobStatus, TrailblazerStatus
from trailblazer.dto.create_analysis_request import CreateAnalysisRequest
from trailblazer.dto.create_job_request import CreateJobRequest
from trailblazer.store.base import BaseHandler
from trailblazer.store.database import get_session
from trailblazer.store.models import Analysis, Job, User


class CreateHandler(BaseHandler):
    """Class for creating items in the database."""

    def add_pending_analysis(self, analysis_data: CreateAnalysisRequest) -> Analysis:
        session: Session = get_session()

        new_analysis = Analysis(
            config_path=analysis_data.config_path,
            workflow=analysis_data.workflow,
            case_id=analysis_data.case_id,
            out_dir=analysis_data.out_dir,
            order_id=analysis_data.order_id,
            priority=analysis_data.priority,
            started_at=datetime.now(),
            status=TrailblazerStatus.PENDING,
            ticket_id=analysis_data.ticket,
            type=analysis_data.type,
            workflow_manager=analysis_data.workflow_manager,
        )

        if analysis_data.email:
            user: User = self.get_user(analysis_data.email)
            new_analysis.user = user

        session.add(new_analysis)
        session.commit()
        return new_analysis

    def add_user(self, name: str, email: str) -> User:
        """Add a new user to the database."""
        session: Session = get_session()
        new_user = User(email=email, name=name)
        session.add(new_user)
        session.commit()
        return new_user

    def add_job(self, analysis_id: int, job_request: CreateJobRequest) -> Job:
        """Add a new job to the database."""
        analysis: Analysis = self.get_analysis_with_id(analysis_id)
        job = Job(
            status=SlurmJobStatus.PENDING,
            slurm_id=job_request.slurm_id,
            job_type=job_request.job_type,
        )
        analysis.jobs.append(job)
        session: Session = get_session()
        session.commit()
        return job

    def replace_jobs(self, analysis_id: int, jobs: list[Job]):
        analysis: Analysis = self.get_analysis_with_id(analysis_id)
        self.delete_analysis_jobs(analysis)
        for job in jobs:
            analysis.jobs.append(job)
        session: Session = get_session()
        session.commit()



from datetime import datetime, timedelta
from pathlib import Path
from typing import Dict, Generator

import pytest
from sqlalchemy.orm import Session

from tests.apps.tower.conftest import TOWER_ID, CaseId, TowerResponseFile, TowerTaskResponseFile
from tests.mocks.store_mock import MockStore
from tests.store.utils.store_helper import StoreHelpers
from trailblazer.apps.tower.models import TowerTask
from trailblazer.constants import (
    TOWER_TIMESTAMP_FORMAT,
    FileExtension,
    FileFormat,
    JobType,
    TrailblazerStatus,
)
from trailblazer.io.controller import ReadFile
from trailblazer.store.database import (
    create_all_tables,
    drop_all_tables,
    get_session,
    initialize_database,
)
from trailblazer.store.models import Analysis, Job
from trailblazer.store.store import Store


@pytest.fixture(scope="session")
def analysis_id_does_not_exist() -> int:
    return 123456789666


@pytest.fixture(scope="session")
def analysis_comment() -> str:
    """Return a comment."""
    return "a comment"


@pytest.fixture(scope="session")
def tower_id() -> str:
    """Return a NF Tower id."""
    return TOWER_ID


@pytest.fixture(scope="session")
def username() -> str:
    """Return a username."""
    return "Paul Anderson"


@pytest.fixture(scope="session")
def archived_username() -> str:
    """Return an archived username."""
    return "Archived User"


@pytest.fixture(scope="session")
def archived_user_email() -> str:
    """Return an archived user email."""
    return "archived.user@magnolia.com"


@pytest.fixture(scope="session")
def user_email() -> str:
    """Return an user email."""
    return "paul.anderson@magnolia.com"


@pytest.fixture(scope="session")
def fixtures_dir() -> Path:
    """Return the path to the fixtures' dir."""
    return Path("tests", "fixtures")


@pytest.fixture(scope="session")
def squeue_dir(fixtures_dir: Path) -> Path:
    """Return the path to the squeue fixture directory."""
    return Path(fixtures_dir, "squeue")


@pytest.fixture(scope="session")
def analysis_data_path(fixtures_dir: Path) -> Path:
    """Return the path to an analysis data file."""
    return Path(fixtures_dir, "analysis-data.yaml")


@pytest.fixture(scope="function")
def analysis_data(analysis_data_path: Path) -> dict[str, list]:
    """Return content of the analysis data file."""
    return ReadFile.get_content_from_file(file_format=FileFormat.YAML, file_path=analysis_data_path)


@pytest.fixture(scope="session")
def squeue_stream_jobs() -> str:
    """Return a squeue output stream."""
    return """JOBID,NAME,STATE,TIME_LIMIT,TIME,START_TIME
690994,gatk_genotypegvcfs2,COMPLETED,10:00:00,1:01:52,2020-10-22T11:43:33
702463,bwa_mem_ACC3218A1,COMPLETED,1-06:00:00,1-1:28:12,2020-10-27T23:06:34
690992,gatk_genotypegvcfs3,COMPLETED,10:00:00,5:54,2020-10-22T11:43:02
690988,gatk_genotypegvcfs4,RUNNING,10:00:00,0:19,N/A
690989,gatk_genotypegvcfs5,PENDING,10:00:00,0:00,N/A"""


@pytest.fixture
def trailblazer_tmp_dir(tmpdir_factory) -> Path:
    """Return a temporary directory for Trailblazer testing."""
    return tmpdir_factory.mktemp("trailblazer_tmp")


@pytest.fixture
def trailblazer_context(analysis_store: MockStore) -> dict[str, MockStore]:
    """Trailblazer context to be used in CLI."""
    return {"trailblazer_db": analysis_store}


@pytest.fixture
def store() -> Generator[MockStore, None, None]:
    """Empty Trailblazer database."""
    initialize_database("sqlite:///:memory:")
    _store = Store()
    create_all_tables()
    yield _store
    drop_all_tables()


@pytest.fixture(scope="function")
def info_store(store: MockStore) -> Generator[MockStore, None, None]:
    """A Trailblazer database wih a populated info table."""
    StoreHelpers.add_info(store=store)
    yield store


@pytest.fixture(scope="function")
def job_store(
    store: MockStore,
) -> Generator[MockStore, None, None]:
    """A Trailblazer database wih a populated job table."""
    statuses: list[str] = [TrailblazerStatus.COMPLETED, TrailblazerStatus.FAILED]
    for index, status in enumerate(statuses):
        StoreHelpers.add_job(analysis_id=index, name=str(index), slurm_id=index, status=status)
    yield store


@pytest.fixture(scope="function")
def user_store(
    archived_user_email: str,
    archived_username: str,
    user_email: str,
    username: str,
    store: MockStore,
) -> Generator[MockStore, None, None]:
    """A Trailblazer database wih a populated user table."""
    StoreHelpers.add_user(email=user_email, name=username)
    StoreHelpers.add_user(email=archived_user_email, name=archived_username, is_archived=True)
    yield store


@pytest.fixture(scope="function")
def raw_analyses(analysis_data: dict[str, list[Dict]]) -> list[dict]:
    """Return raw analyses data."""
    analyses: list[dict] = []
    for analysis in analysis_data["analyses"]:
        analyses.append(analysis)
    return analyses


@pytest.fixture
def analysis_store(
    analysis_data: dict[str, list],
    archived_user_email: str,
    archived_username: str,
    raw_analyses: list[dict],
    store: MockStore,
) -> Generator[MockStore, None, None]:
    """A sample Trailblazer database populated with pending analyses."""
    session: Session = get_session()
    StoreHelpers.add_user(email=archived_user_email, name=archived_username, is_archived=True)
    for user_data in analysis_data["users"]:
        store.add_user(name=user_data["name"], email=user_data["email"])
    for raw_analysis in raw_analyses:
        raw_analysis["user"] = store.get_user(email=raw_analysis["user"])
        raw_analysis["case_id"] = raw_analysis.pop("case_id")
        session.add(Analysis(**raw_analysis))
    yield store


@pytest.fixture
def balsamic_case_id() -> str:
    return "lateraligator"


@pytest.fixture
def balsamic_qc_case_id() -> str:
    return "assumedcrocodile"


@pytest.fixture
def mip_dna_case_id() -> str:
    return "escapedgoat"


@pytest.fixture(scope="session")
def timestamp_now() -> datetime:
    """Return a time stamp of today's date in date time format."""
    return datetime.now()


@pytest.fixture(scope="session")
def timestamp_yesterday(timestamp_now: datetime) -> datetime:
    """Return a time stamp of yesterday's date in date time format."""
    return timestamp_now - timedelta(days=1)


@pytest.fixture(scope="session")
def timestamp_old() -> datetime:
    """Return an old time stamp in a date time format."""
    timestamp: str = "1066-10-14 00:00:00"
    timestamp_format: str = "%Y-%m-%d %H:%M:%S"
    return datetime.strptime(timestamp, timestamp_format)


@pytest.fixture
def tower_jobs(analysis_id, started_at, slurm_id, tower_task_name) -> list[dict]:
    """Return a list of Tower Jobs."""
    return [
        dict(
            analysis_id=analysis_id,
            slurm_id=slurm_id,
            name=tower_task_name,
            started_at=started_at,
            elapsed=63,
            status=TrailblazerStatus.COMPLETED,
        ),
        dict(
            analysis_id=analysis_id,
            slurm_id="4611829",
            name="NFCORE_RNAFUSION:RNAFUSION:PIZZLY_WORKFLOW:KALLISTO_QUANT",
            started_at=None,
            elapsed=0,
            status=TrailblazerStatus.PENDING,
        ),
        dict(
            analysis_id=analysis_id,
            slurm_id="4611828",
            name="NFCORE_RNAFUSION:RNAFUSION:FASTQC",
            started_at=None,
            elapsed=0,
            status=TrailblazerStatus.PENDING,
        ),
    ]


@pytest.fixture
def analysis_id() -> int:
    """Return a mock ID of the analysis in the Trailblazer database."""
    return 1


@pytest.fixture(scope="session")
def started_at() -> datetime:
    """Returns a started at date."""
    return datetime.strptime("2023-04-04T08:11:27Z", TOWER_TIMESTAMP_FORMAT)


@pytest.fixture(scope="session")
def slurm_id() -> str:
    """Returns a slurm id."""
    return "4611827"


@pytest.fixture(scope="session")
def tower_task_name() -> str:
    """Returns a NF Tower task name."""
    return "NFCORE_RNAFUSION:RNAFUSION:INPUT_CHECK:SAMPLESHEET_CHECK"


@pytest.fixture(scope="session")
def case_id() -> str:
    """Return a case id."""
    return CaseId.RUNNING


@pytest.fixture(scope="session")
def case_id_not_in_db() -> str:
    """Return a case id not present in database."""
    return "case_id_not_in_db"


@pytest.fixture(scope="session")
def failed_analysis_case_id() -> str:
    """Return a case id for a failed analysis."""
    return "crackpanda"


@pytest.fixture(scope="session")
def ongoing_analysis_case_id() -> str:
    """Return a case id for an ongoing analysis."""
    return "blazinginsect"


@pytest.fixture(scope="session")
def tower_task() -> TowerTask:
    """Return a Tower task."""
    tower_task_running_content: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=TowerTaskResponseFile.RUNNING
    )
    return TowerTask(**tower_task_running_content["tasks"][0]["task"])


@pytest.fixture(scope="session")
def slurm_squeue_output(squeue_dir: Path) -> dict[str, str]:
    """Return SLURM squeue output for analysis started via SLURM."""
    file_postfix: str = f"squeue{FileExtension.CSV}"
    case_ids: list[str] = [
        "blazinginsect",
        "crackpanda",
        "cuddlyhen",
        "daringpidgeon",
        "escapedgoat",
        "fancymole",
        "happycow",
        "lateraligator",
        "liberatedunicorn",
        "nicemice",
        "rarekitten",
        "trueferret",
    ]
    return {
        case_id: Path(squeue_dir, f"{case_id}_{file_postfix}").as_posix() for case_id in case_ids
    }


@pytest.fixture
def slurm_job_ids() -> list[int]:
    return [690993, 690994, 690992, 690988, 690989, 690990]


@pytest.fixture(scope="session")
def tower_case_config() -> dict[str, dict]:
    """Return a Tower case configs."""
    return {
        CaseId.RUNNING: {
            "workflow_response_file": TowerResponseFile.RUNNING,
            "tasks_response_file": TowerTaskResponseFile.RUNNING,
            "tower_id": TOWER_ID,
            "analysis_id": 1,
        },
        CaseId.PENDING: {
            "workflow_response_file": TowerResponseFile.PENDING,
            "tasks_response_file": TowerTaskResponseFile.PENDING,
            "tower_id": TOWER_ID,
            "analysis_id": 1,
        },
        CaseId.COMPLETED: {
            "workflow_response_file": TowerResponseFile.COMPLETED,
            "tasks_response_file": TowerTaskResponseFile.COMPLETED,
            "tower_id": TOWER_ID,
            "analysis_id": 1,
        },
    }


@pytest.fixture
def analysis_with_upload_jobs(case_id: str) -> Analysis:
    upload_job = Job(
        name="upload",
        slurm_id="123456",
        job_type=JobType.UPLOAD,
        status=TrailblazerStatus.COMPLETED,
        started_at=datetime.now(),
        elapsed=10,
    )
    analysis = Analysis(case_id=case_id)
    analysis.jobs.append(upload_job)
    session = get_session()
    session.add(analysis)
    return analysis



from functools import partial

import pytest
from click.testing import CliRunner

from trailblazer.cli.core import base

from trailblazer.containers import Container
from trailblazer.server.wiring import setup_dependency_injection


@pytest.fixture(scope="session")
def process_exit_success() -> int:
    """Return the exit code for a successful process."""
    return 0


@pytest.fixture(scope="session")
def cli_runner() -> CliRunner:
    """Return a CliRunner fixture."""
    return CliRunner()


@pytest.fixture(scope="session")
def invoke_cli(cli_runner: CliRunner) -> partial:
    """Invokes CLI base with partial functionality."""
    return partial(cli_runner.invoke, base)


@pytest.fixture(autouse=True)
def container() -> Container:
    container: Container = setup_dependency_injection()
    return container



from unittest.mock import Mock
import pytest

from trailblazer.services.analysis_service.analysis_service import AnalysisService
from trailblazer.services.job_service.job_service import JobService
from trailblazer.services.slurm.dtos import SlurmJobInfo
from trailblazer.services.slurm.slurm_cli_service.slurm_cli_service import SlurmCLIService
from trailblazer.services.slurm.slurm_service import SlurmService
from trailblazer.store.store import Store


@pytest.fixture
def slurm_completed_job_info() -> SlurmJobInfo:
    return SlurmJobInfo(
        slurm_id=690994,
        name="job",
        status="completed",
        elapsed=1000,
        started_at="2020-10-22T11:43:33",
    )


@pytest.fixture
def slurm_service(slurm_completed_job_info) -> SlurmService:
    """Slurm service reporting all jobs as completed."""
    service = SlurmCLIService(client=Mock())
    service.get_job = Mock(return_value=slurm_completed_job_info)
    service.get_jobs = Mock(return_value=[slurm_completed_job_info])
    return service


@pytest.fixture
def job_service(store: Store, slurm_service: SlurmService, slurm_job_ids, mocker) -> JobService:
    mocker.patch(
        "trailblazer.services.job_service.job_service.get_slurm_job_ids", return_value=slurm_job_ids
    )
    return JobService(store=store, slurm_service=slurm_service)


@pytest.fixture
def analysis_service(analysis_store: Store, job_service: JobService) -> AnalysisService:
    return AnalysisService(store=analysis_store, job_service=job_service)



from datetime import datetime

from trailblazer.apps.slurm.models import SqueueJob, SqueueResult
from trailblazer.constants import FileFormat, SlurmJobStatus
from trailblazer.io.controller import ReadStream

LONG_RUNNING_JOB: int = 1528
MEDIUM_RUNNING_JOOB: int = 61
SHORT_RUNNING_JOB: int = 5


def test_instantiate_squeue_result(squeue_stream_pending_job: str):
    """Tests squeue output against a pydantic model."""
    # GIVEN a csv squeue stream

    csv_content: list[dict] = ReadStream.get_content_from_stream(
        file_format=FileFormat.CSV,
        stream=squeue_stream_pending_job,
        read_to_dict=True,
    )

    # WHEN instantiating a SqueueJob object
    squeue_result: SqueueResult = SqueueResult(jobs=csv_content)

    # THEN it should be successfully created
    assert isinstance(squeue_result, SqueueResult)


def test_instantiate_squeue_job(squeue_stream_pending_job: str):
    """Tests squeue row output against a pydantic model."""
    # GIVEN a csv squeue stream

    csv_content: list[dict] = ReadStream.get_content_from_stream(
        file_format=FileFormat.CSV,
        stream=squeue_stream_pending_job,
        read_to_dict=True,
    )
    for job in csv_content:
        # WHEN instantiating a SqueueJob object
        squeue_job: SqueueJob = SqueueJob(**job)

        # THEN it was successfully created
        assert isinstance(squeue_job, SqueueJob)


def test_squeue_job_convert_time_elapse_to_minutes(squeue_stream_jobs: str):
    """Tests SqueueJob convert time elapsed to minutes validator."""
    # GIVEN a csv squeue stream

    csv_content: list[dict] = ReadStream.get_content_from_stream(
        file_format=FileFormat.CSV,
        stream=squeue_stream_jobs,
        read_to_dict=True,
    )
    expected_time_elapsed: dict[int, int] = {
        0: MEDIUM_RUNNING_JOOB,
        1: LONG_RUNNING_JOB,
        2: SHORT_RUNNING_JOB,
        3: 0,
        4: 0,
    }
    for index, job in enumerate(csv_content):
        # WHEN instantiating a SqueueJob object
        squeue_job: SqueueJob = SqueueJob(**job)

        # THEN the time elapsed should be in min
        assert squeue_job.time_elapsed == expected_time_elapsed[index]


def test_convert_status_to_lower_case(squeue_stream_pending_job: str):
    """Tests converting status to lower case."""
    # GIVEN a csv squeue stream

    csv_content: list[dict] = ReadStream.get_content_from_stream(
        file_format=FileFormat.CSV,
        stream=squeue_stream_pending_job,
        read_to_dict=True,
    )
    for job in csv_content:
        # WHEN instantiating a SqueueJob object
        squeue_job: SqueueJob = SqueueJob(**job)

        # THEN status is in lower case
        assert squeue_job.status.islower()


def test_convert_started_to_datetime(squeue_stream_pending_job: str):
    """Tests converting started to datetime."""
    # GIVEN a csv squeue stream

    csv_content: list[dict] = ReadStream.get_content_from_stream(
        file_format=FileFormat.CSV,
        stream=squeue_stream_pending_job,
        read_to_dict=True,
    )
    for job in csv_content:
        # WHEN instantiating a SqueueJob object
        squeue_job: SqueueJob = SqueueJob(**job)

        # THEN started at is a datetime object
        assert isinstance(squeue_job.started_at, datetime)


def test_convert_started_to_datetime_no_datetime_format(squeue_stream_pending_job_not_started: str):
    """Tests converting started to datetime when started is not in datetime format."""
    # GIVEN a csv squeue stream

    csv_content: list[dict] = ReadStream.get_content_from_stream(
        file_format=FileFormat.CSV,
        stream=squeue_stream_pending_job_not_started,
        read_to_dict=True,
    )
    for job in csv_content:
        # WHEN instantiating a SqueueJob object
        squeue_job: SqueueJob = SqueueJob(**job)

        # THEN started at is not initialized
        assert not squeue_job.started_at


def test_set_jobs_status_distribution(squeue_stream_pending_job: str):
    """Tests set job status distribution from jobs status."""
    # GIVEN a csv squeue stream
    csv_content: list[dict] = ReadStream.get_content_from_stream(
        file_format=FileFormat.CSV,
        stream=squeue_stream_pending_job,
        read_to_dict=True,
    )

    # WHEN instantiating a SqueueResult object
    squeue_result: SqueueResult = SqueueResult(jobs=csv_content)

    # THEN a job status distribution should be returned
    assert squeue_result.jobs_status_distribution == {SlurmJobStatus.PENDING: 1}



from trailblazer.store.models import User


def test_user_first_name():
    """Test setting user first name."""
    # GIVEN a user with a name
    user: User = User(name="Paul T. Anderson")

    # WHEN accessing the first name
    first_name = user.first_name

    # THEN it should return the spoken name
    assert first_name == "Paul"



from abc import ABC, abstractmethod

from trailblazer.services.slurm.dtos import SlurmJobInfo


class SlurmService(ABC):
    @abstractmethod
    def get_job(self, job_id: int) -> SlurmJobInfo:
        pass

    @abstractmethod
    def get_jobs(self, job_ids: list[int]) -> list[SlurmJobInfo]:
        pass



from pydantic import BaseModel

from trailblazer.constants import TrailblazerStatus


class AnalysisUpdateRequest(BaseModel):
    comment: str | None = None
    is_delivered: bool | None = None
    is_visible: bool | None = None
    status: TrailblazerStatus | None = None



import requests

from trailblazer.clients.google_api_client.exceptions import GoogleAPIClientError


class GoogleAPIClient:
    def __init__(self, base_url: str):
        self.base_url = base_url

    def _get_headers(self, access_token: str) -> dict:
        return {"Authorization": f"Bearer {access_token}"}

    def get_user_email(self, access_token: str) -> str:
        """Get the user email for the given access token."""
        endpoint: str = f"{self.base_url}/oauth2/v1/userinfo"
        headers: dict = self._get_headers(access_token)
        response = requests.get(endpoint, headers=headers)

        if not response.ok:
            raise GoogleAPIClientError(response.text)

        return response.json()["email"]



from http import HTTPStatus

from flask.testing import FlaskClient

from trailblazer.store.models import Analysis


def test_get_summaries(client: FlaskClient, analysis: Analysis):
    # GIVEN an order with an analysis
    order_id: int = analysis.order_id

    # WHEN requesting a summary for the analyses in the order
    response = client.get(f"/api/v1/summary?orderIds={order_id}")

    # THEN it gives a success response
    assert response.status_code == HTTPStatus.OK


def test_get_summaries_lastest_date(
    client: FlaskClient,
    analysis: Analysis,
    analyses: list[Analysis],
    order_id_with_multiple_analyses: int,
):
    # GIVEN an order with multiple analyses for the same case

    # WHEN requesting a summary for the analyses in the order
    response = client.get(f"/api/v1/summary?orderIds={order_id_with_multiple_analyses}")

    # THEN it gives a success response
    assert response.status_code == HTTPStatus.OK
    assert response.json["summaries"][0]["total"] == 1



from http import HTTPStatus

from flask.testing import FlaskClient

from trailblazer.dto.update_analyses import AnalysisUpdate, UpdateAnalyses
from trailblazer.store.models import Analysis


def test_patch_analysis(client: FlaskClient, analysis: Analysis):
    # GIVEN an analysis

    # GIVEN a valid request to pach the analysis
    update = AnalysisUpdate(id=analysis.id, comment="new_comment")
    request = UpdateAnalyses(analyses=[update])
    data: str = request.model_dump_json()

    # WHEN patching the analysis
    response = client.patch("/api/v1/analyses", data=data, content_type="application/json")

    # THEN it gives a success response
    assert response.status_code == HTTPStatus.OK

    # THEN it should return the analysis
    assert response.json["analyses"]



from datetime import datetime
from enum import Enum
from typing import Callable

import sqlalchemy
from sqlalchemy import Subquery, func, select
from sqlalchemy.orm import Query

from trailblazer.constants import TrailblazerStatus, Workflow
from trailblazer.dto.analyses_request import AnalysisSortField
from trailblazer.dto.common import SortOrder
from trailblazer.store.database import get_session
from trailblazer.store.models import Analysis


def filter_analyses_by_comment(analyses: Query, comment: str, **kwargs) -> Query:
    """Filter analyses that contain the string given in 'comment'."""
    return analyses.filter(Analysis.comment.ilike(f"%{comment}%"))


def filter_analyses_by_case_id(analyses: Query, case_id: str | None, **kwargs) -> Query:
    """Filter analyses by case id."""
    if case_id is None:
        return analyses
    return analyses.filter(Analysis.case_id == case_id)


def filter_analyses_by_entry_id(analyses: Query, analysis_id: int, **kwargs) -> Query:
    """Filter analyses by database entry id."""
    return analyses.filter(Analysis.id == analysis_id)


def filter_analyses_by_is_visible(analyses: Query, show_hidden: bool | None, **kwargs) -> Query:
    return analyses if show_hidden else analyses.filter(Analysis.is_visible.is_(True))


def filter_analyses_by_order_id(analyses: Query, order_id: int | None, **kwargs) -> Query:
    if order_id is None:
        return analyses
    return analyses.filter(Analysis.order_id == order_id)


def filter_analyses_by_search_term(analyses: Query, search_term: str | None, **kwargs) -> Query:
    """Filter analyses by search term using multiple fields."""
    if not search_term:
        return analyses
    return analyses.filter(
        sqlalchemy.or_(
            Analysis.case_id.ilike(f"%{search_term}%"),
            Analysis.status.ilike(f"%{search_term}%"),
            Analysis.workflow.ilike(f"%{search_term}%"),
            Analysis.comment.ilike(f"%{search_term}%"),
        )
    )


def filter_analyses_by_started_at(analyses: Query, started_at: datetime, **kwargs) -> Query:
    """Filter analyses by when it started."""
    return analyses.filter(Analysis.started_at == started_at)


def filter_analyses_by_before_started_at(analyses: Query, started_at: datetime, **kwargs) -> Query:
    """Filter analyses by ifK it started before the supplied date."""
    return analyses.filter(Analysis.started_at < started_at)


def filter_analyses_by_status(analyses: Query, status: TrailblazerStatus, **kwargs) -> Query:
    """Filter analyses by status."""
    return analyses.filter(Analysis.status == status)


def filter_analyses_by_statuses(analyses: Query, statuses: list[str] | None, **kwargs) -> Query:
    return analyses.filter(Analysis.status.in_(statuses)) if statuses else analyses


def filter_analyses_by_priorites(analyses: Query, priorities: list[str] | None, **kwargs) -> Query:
    return analyses.filter(Analysis.priority.in_(priorities)) if priorities else analyses


def filter_analyses_by_types(analyses: Query, types: list[str] | None, **kwargs) -> Query:
    return analyses.filter(Analysis.type.in_(types)) if types else analyses


def filter_analyses_by_has_comment(analyses: Query, has_comment: bool | None, **kwargs) -> Query:
    if has_comment is None:
        return analyses
    if has_comment:
        return analyses.filter(
            sqlalchemy.and_(Analysis.comment.isnot(None), Analysis.comment != "")
        )
    return analyses.filter(sqlalchemy.or_(Analysis.comment.is_(None), Analysis.comment == ""))


def filter_analyses_by_delivered(analyses: Query, delivered: bool | None, **kwargs) -> Query:
    if delivered is True:
        return analyses.filter(Analysis.delivery != None)
    elif delivered is False:
        return analyses.filter(Analysis.delivery == None)
    return analyses


def filter_analyses_by_workflow(analyses: Query, workflow: Workflow, **kwargs) -> Query:
    """Filter analyses by workflow."""
    balsamic_workflow: str = Workflow.BALSAMIC.lower()
    if workflow == balsamic_workflow:
        analyses = analyses.filter(Analysis.workflow.startswith(balsamic_workflow))
    elif workflow:
        analyses = analyses.filter(Analysis.workflow == workflow)
    return analyses


def filter_analyses_by_latest_per_case(analyses: Query, **kwargs) -> Query:
    """Return only the latest analyses per case.

    Constructs a query which reads as:
        SELECT
            *
        FROM
            Analysis
        JOIN
            (SELECT
                case_id, MAX(started_at) started_at
            FROM
                analyses
            GROUP BY
                case_id) latest_started_analyses
        USING (case_id, started_at)
    """

    provided_analyses = analyses.subquery()
    session = get_session()
    latest_date_per_case: Subquery = (
        session.query(
            provided_analyses.columns.case_id,
            func.max(provided_analyses.columns.started_at).label("max_started_at"),
        )
        .group_by(provided_analyses.columns.case_id)
        .subquery()
    )
    latest_analysis_per_case: Query = analyses.join(
        latest_date_per_case,
        (Analysis.case_id == latest_date_per_case.c.case_id)
        & (Analysis.started_at == latest_date_per_case.c.max_started_at),
    )

    return latest_analysis_per_case


def sort_analyses(
    analyses: Query, sort_field: AnalysisSortField, sort_order: SortOrder, **kwargs
) -> Query:
    if not sort_field or not sort_order:
        return analyses
    column = getattr(Analysis, sort_field)
    if sort_order == SortOrder.ASC:
        return analyses.order_by(sqlalchemy.asc(column))
    return analyses.order_by(sqlalchemy.desc(column))


def paginate_analyses(analyses: Query, page: int, page_size: int, **kwargs) -> Query:
    if not page or not page_size:
        return analyses
    return analyses.limit(page_size).offset((page - 1) * page_size)


def exclude_rsync_analyses(analyses: Query, **kwargs) -> Query:
    return analyses.filter(Analysis.workflow != Workflow.RSYNC)


class AnalysisFilter(Enum):
    """Define Analysis filter functions."""

    BY_BEFORE_STARTED_AT: Callable = filter_analyses_by_before_started_at
    BY_CASE_ID: Callable = filter_analyses_by_case_id
    BY_COMMENT: Callable = filter_analyses_by_comment
    BY_HAS_COMMENT: Callable = filter_analyses_by_has_comment
    BY_ENTRY_ID: Callable = filter_analyses_by_entry_id
    BY_SEARCH_TERM: Callable = filter_analyses_by_search_term
    BY_IS_VISIBLE: Callable = filter_analyses_by_is_visible
    BY_ORDER_ID: Callable = filter_analyses_by_order_id
    BY_PRIORITIES: Callable = filter_analyses_by_priorites
    BY_STARTED_AT: Callable = filter_analyses_by_started_at
    BY_STATUS: Callable = filter_analyses_by_status
    BY_STATUSES: Callable = filter_analyses_by_statuses
    BY_TYPES: Callable = filter_analyses_by_types
    BY_DELIVERED: Callable = filter_analyses_by_delivered
    BY_WORKFLOW: Callable = filter_analyses_by_workflow
    BY_LATEST_PER_CASE: Callable = filter_analyses_by_latest_per_case
    EXCLUDE_RSYNC: Callable = exclude_rsync_analyses
    SORTING: Callable = sort_analyses
    PAGINATION: Callable = paginate_analyses


def apply_analysis_filter(
    analyses: Query,
    filter_functions: list[Callable],
    analysis_id: int | None = None,
    case_id: str | None = None,
    comment: str | None = None,
    order_id: int | None = None,
    priorities: list[str] | None = None,
    search_term: str | None = None,
    started_at: datetime | None = None,
    status: str | None = None,
    statuses: list[str] | None = None,
    types: list[str] | None = None,
    delivered: bool | None = None,
    workflow: str | None = None,
    has_comment: bool | None = None,
    show_hidden: bool | None = None,
    page: int | None = None,
    page_size: int | None = None,
    sort_field: AnalysisSortField | None = None,
    sort_order: SortOrder | None = None,
) -> Query:
    """Apply filtering functions and return filtered results."""
    if statuses is None:
        statuses = []
    for function in filter_functions:
        analyses: Query = function(
            analyses=analyses,
            analysis_id=analysis_id,
            case_id=case_id,
            comment=comment,
            order_id=order_id,
            priorities=priorities,
            search_term=search_term,
            started_at=started_at,
            status=status,
            statuses=statuses,
            types=types,
            delivered=delivered,
            workflow=workflow,
            has_comment=has_comment,
            show_hidden=show_hidden,
            page=page,
            page_size=page_size,
            sort_field=sort_field,
            sort_order=sort_order,
        )
    return analyses



import os
from dependency_injector import containers, providers

from trailblazer.clients.authentication_client.google_oauth_client import GoogleOAuthClient
from trailblazer.clients.google_api_client.google_api_client import GoogleAPIClient
from trailblazer.clients.slurm_cli_client.slurm_cli_client import SlurmCLIClient
from trailblazer.services.analysis_service.analysis_service import AnalysisService
from trailblazer.services.authentication_service.authentication_service import AuthenticationService
from trailblazer.services.encryption_service.encryption_service import EncryptionService
from trailblazer.services.job_service import JobService
from trailblazer.services.slurm.slurm_cli_service.slurm_cli_service import SlurmCLIService
from trailblazer.store.store import Store


class Container(containers.DeclarativeContainer):
    slurm_host: str | None = os.environ.get("ANALYSIS_HOST")
    oauth_client_id: str = os.environ.get("GOOGLE_CLIENT_ID")
    oauth_client_secret: str = os.environ.get("GOOGLE_CLIENT_SECRET")
    oauth_redirect_uri: str = os.environ.get("GOOGLE_REDIRECT_URI")
    google_oauth_base_url: str = os.environ.get("GOOGLE_OAUTH_BASE_URL")
    encryption_key: str = os.environ.get("ENCRYPTION_KEY")
    google_api_base_url: str = os.environ.get("GOOGLE_API_BASE_URL")

    google_api_client = GoogleAPIClient(google_api_base_url)

    google_oauth_client = providers.Singleton(
        GoogleOAuthClient,
        client_id=oauth_client_id,
        client_secret=oauth_client_secret,
        redirect_uri=oauth_redirect_uri,
        oauth_base_url=google_oauth_base_url,
    )

    store = providers.Singleton(Store)
    slurm_client = providers.Singleton(SlurmCLIClient, host=slurm_host)
    slurm_service = providers.Singleton(SlurmCLIService, client=slurm_client)

    job_service = providers.Factory(JobService, store=store, slurm_service=slurm_service)
    analysis_service = providers.Factory(AnalysisService, store=store, job_service=job_service)

    encryption_service = providers.Singleton(EncryptionService, secret_key=encryption_key)

    auth_service = providers.Singleton(
        AuthenticationService,
        google_oauth_client=google_oauth_client,
        google_api_client=google_api_client,
        encryption_service=encryption_service,
        store=store,
    )



from pathlib import Path

from trailblazer.io.json import read_json


def test_get_content_from_file(example_json_path: Path):
    """
    Tests read JSON content from file.
    """
    # GIVEN a JSON file

    # WHEN reading the JSON file
    raw_json_content: dict = read_json(file_path=example_json_path)

    # Then assert a dict is returned
    assert isinstance(raw_json_content, dict)

    # THEN assert the content is unchanged
    assert raw_json_content == {"glossary": {"title": "example glossary"}}



import logging
import os
from http import HTTPStatus
from typing import Mapping

from dependency_injector.wiring import Provide, inject
from flask import Blueprint, Response, abort, g, jsonify, make_response, request
from google.auth import jwt
from pydantic import ValidationError

from trailblazer.containers import Container
from trailblazer.dto import (
    AnalysesRequest,
    AnalysesResponse,
    AnalysisResponse,
    AnalysisUpdateRequest,
    CreateJobRequest,
    FailedJobsRequest,
    FailedJobsResponse,
    JobResponse,
)
from trailblazer.dto.analyses_response import UpdateAnalysesResponse
from trailblazer.dto.authentication.code_exchange_request import CodeExchangeRequest
from trailblazer.dto.create_analysis_request import CreateAnalysisRequest
from trailblazer.dto.summaries_request import SummariesRequest
from trailblazer.dto.summaries_response import SummariesResponse
from trailblazer.dto.update_analyses import UpdateAnalyses
from trailblazer.exc import MissingAnalysis
from trailblazer.server.ext import store
from trailblazer.server.utils import parse_analyses_request, stringify_timestamps
from trailblazer.services.analysis_service.analysis_service import AnalysisService
from trailblazer.services.authentication_service.authentication_service import AuthenticationService
from trailblazer.services.authentication_service.exceptions import AuthenticationError
from trailblazer.services.job_service import JobService
from trailblazer.store.models import Info, User

blueprint = Blueprint("api", __name__, url_prefix="/api/v1")


@blueprint.before_request
def before_request():
    """Authentication that is run before processing requests to the application"""
    if request.endpoint == "api.authenticate":
        return
    if request.method == "OPTIONS":
        return make_response(jsonify(ok=True), 204)
    if os.environ.get("SCOPE") == "DEVELOPMENT":
        return
    if auth_header := request.headers.get("Authorization"):
        jwt_token = auth_header.split("Bearer ")[-1]
    else:
        return abort(403, "no JWT token found on request")

    user_data: Mapping = jwt.decode(jwt_token, verify=False)
    if user := store.get_user(email=user_data["email"], exclude_archived=True):
        g.current_user = user
    else:
        return abort(403, f"{user_data['email']} doesn't have access")


@blueprint.route("/auth", methods=["POST"])
@inject
def authenticate(auth_service: AuthenticationService = Provide[Container.auth_service]):
    """Exchange authorization code for an access token."""
    try:
        request_data = CodeExchangeRequest.model_validate(request.json)
        token: str = auth_service.authenticate(request_data.code)
        return jsonify({"token": token}), HTTPStatus.OK
    except ValidationError as error:
        return jsonify(error=str(error)), HTTPStatus.BAD_REQUEST
    except AuthenticationError:
        return jsonify("User not allowed"), HTTPStatus.FORBIDDEN


@blueprint.route("/auth/refresh", methods=["GET"])
@inject
def refresh_token(auth_service: AuthenticationService = Provide[Container.auth_service]):
    """Refresh access token."""
    user: User = g.current_user
    try:
        token: str = auth_service.refresh_token(user.id)
        return jsonify({"token": token}), HTTPStatus.OK
    except AuthenticationError:
        return jsonify("User not allowed"), HTTPStatus.FORBIDDEN


@blueprint.route("/analyses", methods=["GET"])
@inject
def get_analyses(analysis_service: AnalysisService = Provide[Container.analysis_service]):
    try:
        request_data: AnalysesRequest = parse_analyses_request(request)
        response: AnalysesResponse = analysis_service.get_analyses(request_data)
        return jsonify(response.model_dump()), HTTPStatus.OK
    except ValidationError as error:
        return jsonify(error=str(error)), HTTPStatus.BAD_REQUEST


@blueprint.route("/analyses", methods=["PATCH"])
@inject
def patch_analyses(analysis_service: AnalysisService = Provide[Container.analysis_service]):
    """Update data (such as status, visibility, comments etc.) for multiple analyses at once."""
    try:
        request_data = UpdateAnalyses.model_validate(request.json)
        user: User = g.get("current_user")
        response: UpdateAnalysesResponse = analysis_service.update_analyses(
            data=request_data, user=user
        )
        return jsonify(response.model_dump()), HTTPStatus.OK
    except ValidationError as error:
        return jsonify(error=str(error)), HTTPStatus.BAD_REQUEST
    except MissingAnalysis as error:
        return jsonify(error=str(error)), HTTPStatus.NOT_FOUND


@blueprint.route("/analyses/<int:analysis_id>", methods=["GET"])
@inject
def get_analysis(
    analysis_id: int, analysis_service: AnalysisService = Provide[Container.analysis_service]
):
    try:
        response: AnalysisResponse = analysis_service.get_analysis(analysis_id)
        return jsonify(response.model_dump()), HTTPStatus.OK
    except MissingAnalysis as error:
        return jsonify(error=str(error)), HTTPStatus.NOT_FOUND


@blueprint.route("/analysis/<int:analysis_id>/jobs", methods=["POST"])
@inject
def add_job(analysis_id: int, job_service: JobService = Provide[Container.job_service]):
    try:
        data = CreateJobRequest.model_validate(request.json)
        response: JobResponse = job_service.add_job(analysis_id=analysis_id, data=data)
        return jsonify(response.model_dump()), HTTPStatus.CREATED
    except MissingAnalysis as error:
        return jsonify(error=str(error)), HTTPStatus.NOT_FOUND
    except ValidationError as error:
        return jsonify(error=str(error)), HTTPStatus.BAD_REQUEST


@blueprint.route("/analyses/<int:analysis_id>", methods=["PUT"])
@inject
def update_analysis(
    analysis_id: int, analysis_service: AnalysisService = Provide[Container.analysis_service]
):
    try:
        request_data = AnalysisUpdateRequest.model_validate(request.json)
        user: User = g.get("current_user")
        response: AnalysisResponse = analysis_service.update_analysis(
            analysis_id=analysis_id, update=request_data, user=user
        )
        return jsonify(response.model_dump()), HTTPStatus.OK
    except MissingAnalysis as error:
        return jsonify(error=str(error)), HTTPStatus.NOT_FOUND
    except ValidationError as error:
        return jsonify(error=str(error)), HTTPStatus.BAD_REQUEST


@blueprint.route("/summary", methods=["GET"])
@inject
def get_summaries(analysis_service: AnalysisService = Provide[Container.analysis_service]):
    try:
        request_data = SummariesRequest.model_validate(request.args)
        response: SummariesResponse = analysis_service.get_summaries(request_data)
        return jsonify(response.model_dump()), HTTPStatus.OK
    except ValidationError as error:
        return jsonify(error=str(error)), HTTPStatus.BAD_REQUEST


@blueprint.route("/info")
def info():
    """Display metadata about database."""
    info: Info = store.get_query(table=Info).first()
    return jsonify(**info.to_dict())


@blueprint.route("/me")
def me():
    """Return information about a logged in user."""
    return jsonify(**g.current_user.to_dict())


@blueprint.route("/aggregate/jobs", methods=["GET"])
@inject
def get_failed_jobs(job_service: JobService = Provide[Container.job_service]):
    try:
        request_data = FailedJobsRequest.model_validate(request.args)
        response: FailedJobsResponse = job_service.get_failed_jobs(request_data)
        return jsonify(response.model_dump()), HTTPStatus.OK
    except ValidationError as error:
        return jsonify(error=str(error)), HTTPStatus.BAD_REQUEST


# CG REST INTERFACE ###
# ONLY POST routes which accept messages in specific format
# NOT for use with GUI (for now)


@blueprint.route("/get-latest-analysis", methods=["POST"])
def post_get_latest_analysis():
    """Return latest analysis entry for specified case id."""
    post_request = request.json
    case_id: str = post_request.get("case_id")
    if latest_case_analysis := store.get_latest_analysis_for_case(case_id):
        raw_analysis: dict[str, str] = stringify_timestamps(latest_case_analysis.to_dict())
        return jsonify(**raw_analysis), HTTPStatus.OK
    return jsonify(None), HTTPStatus.OK


@blueprint.route("/add-pending-analysis", methods=["POST"])
@inject
def post_add_pending_analysis(
    analysis_service: AnalysisService = Provide[Container.analysis_service],
):
    try:
        request_data = CreateAnalysisRequest.model_validate(request.json)
        response: AnalysisResponse = analysis_service.add_pending_analysis(request_data)
        return jsonify(response.model_dump()), HTTPStatus.CREATED
    except ValidationError as error:
        return jsonify(error=str(error)), HTTPStatus.BAD_REQUEST


@blueprint.route("/set-analysis-uploaded", methods=["PUT"])
def set_analysis_uploaded():
    """Set the analysis uploaded at attribute."""
    put_request: Response.json = request.json
    try:
        store.update_analysis_uploaded_at(
            case_id=put_request.get("case_id"), uploaded_at=put_request.get("uploaded_at")
        )
        return jsonify("Success! Uploaded at request sent"), HTTPStatus.CREATED
    except Exception as error:
        return jsonify(f"Exception: {error}"), HTTPStatus.CONFLICT


@blueprint.route("/set-analysis-status", methods=["PUT"])
def set_analysis_status():
    """Update analysis status of a case with supplied status."""
    put_request: Response.json = request.json
    try:
        case_id: str = put_request.get("case_id")
        status: str = put_request.get("status")
        store.update_analysis_status_by_case_id(case_id=case_id, status=status)
        return (
            jsonify(f"Success! Analysis set to {put_request.get('status')} request sent"),
            HTTPStatus.CREATED,
        )
    except Exception as error:
        return jsonify(f"Exception: {error}"), HTTPStatus.CONFLICT


@blueprint.route("/add-comment", methods=["PUT"])
def add_comment():
    """Updating comment on analysis."""
    put_request: Response.json = request.json
    try:
        case_id: str = put_request.get("case_id")
        comment: str = put_request.get("comment")
        store.update_analysis_comment(case_id=case_id, comment=comment)
        return jsonify("Success! Adding comment request sent"), HTTPStatus.CREATED
    except Exception as error:
        return jsonify(f"Exception: {error}"), HTTPStatus.CONFLICT



from trailblazer.containers import Container

container = Container()


def setup_dependency_injection() -> Container:
    import trailblazer.server.api
    import trailblazer.cli.core

    container.wire(modules=[trailblazer.server.api, trailblazer.cli.core])
    return container



from trailblazer.dto.analyses_request import AnalysesRequest  # noqa: F401
from trailblazer.dto.analyses_response import AnalysesResponse  # noqa: F401
from trailblazer.dto.analysis_response import AnalysisResponse  # noqa: F401
from trailblazer.dto.analysis_update_request import AnalysisUpdateRequest  # noqa: F401
from trailblazer.dto.failed_jobs_request import FailedJobsRequest  # noqa: F401
from trailblazer.dto.failed_jobs_response import FailedJobsResponse  # noqa: F401
from trailblazer.dto.create_job_request import CreateJobRequest  # noqa: F401
from trailblazer.dto.job_response import JobResponse  # noqa: F401
from trailblazer.dto.create_analysis_request import CreateAnalysisRequest  # noqa: F401



from pathlib import Path
from trailblazer.constants import FileFormat, SlurmJobStatus, TrailblazerStatus
from trailblazer.io.controller import ReadFile
from trailblazer.store.models import Job


def get_slurm_job_ids(job_id_file: str) -> list[int]:
    job_id_file_path = Path(job_id_file)
    content: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.YAML, file_path=job_id_file_path
    )
    job_ids: list[int] = []
    for row in content.values():
        [job_ids.append(job_id) for job_id in row]
    return job_ids


def get_status(jobs: list[Job]) -> TrailblazerStatus:
    if has_same_status(jobs):
        return get_single_status(jobs)
    is_running: bool = has_running_jobs(jobs)
    is_failed: bool = has_failed_jobs(jobs)
    if is_failed:
        return TrailblazerStatus.ERROR if is_running else TrailblazerStatus.FAILED
    return TrailblazerStatus.RUNNING if is_running else TrailblazerStatus.CANCELLED


def has_same_status(jobs: list[Job]) -> bool:
    if not jobs:
        return False
    first_status = jobs[0].status
    return all(job.status == first_status for job in jobs)


def get_single_status(jobs: list[Job]) -> SlurmJobStatus:
    single_status = jobs[0].status
    return (
        TrailblazerStatus.FAILED
        if single_status == SlurmJobStatus.TIME_OUT
        else TrailblazerStatus[single_status.upper()]
    )


def has_running_jobs(jobs: list[Job]) -> bool:
    run_statuses = SlurmJobStatus.ongoing_statuses()
    return any(job.status in run_statuses for job in jobs)


def has_failed_jobs(jobs: list[Job]) -> bool:
    fail_statuses = SlurmJobStatus.fail_statuses()
    return any(job.status in fail_statuses for job in jobs)


def get_progress(jobs: list[Job]) -> float:
    total_jobs: int = len(jobs)
    if total_jobs == 0:
        return 0.0
    completed_jobs: int = len([job for job in jobs if job.status == SlurmJobStatus.COMPLETED])
    return completed_jobs / total_jobs



from flask.testing import FlaskClient
from http import HTTPStatus
from trailblazer.dto.create_job_request import CreateJobRequest

from trailblazer.store.models import Analysis
from trailblazer.store.store import Store

TYPE_JSON = "application/json"


def test_add_job_to_analysis(client: FlaskClient, analysis: Analysis, store: Store):
    # GIVEN an analysis
    analysis_id: int = analysis.id

    # GIVEN a valid request to add a job to the analysis
    create_job_request = CreateJobRequest(slurm_id=123)
    data: str = create_job_request.model_dump_json()

    # WHEN sending the request
    response = client.post(
        f"/api/v1/analysis/{analysis_id}/jobs", data=data, content_type=TYPE_JSON
    )

    # THEN it gives a success response
    assert response.status_code == HTTPStatus.CREATED

    # THEN it should return the job
    assert response.json["slurm_id"] == create_job_request.slurm_id

    # THEN the job was persisted
    assert store.get_analysis_with_id(analysis_id).jobs



from http import HTTPStatus

from flask.testing import FlaskClient

from trailblazer.constants import TrailblazerPriority, TrailblazerTypes
from trailblazer.dto.create_analysis_request import CreateAnalysisRequest
from trailblazer.store.store import Store

TYPE_JSON = "application/json"


def test_adding_analysis(client: FlaskClient, store: Store):
    # GIVEN a valid request to add an analysis
    create_analysis_request = CreateAnalysisRequest(
        case_id="case_id",
        config_path="config_path",
        workflow=TrailblazerTypes.WGS,
        priority=TrailblazerPriority.NORMAL,
        out_dir="out_dir",
        ticket="ticket_id",
        type=TrailblazerTypes.WGS,
        order_id=123,
    )
    data: str = create_analysis_request.model_dump_json()

    # WHEN sending the request
    response = client.post("/api/v1/add-pending-analysis", data=data, content_type=TYPE_JSON)

    # THEN it gives a success response
    assert response.status_code == HTTPStatus.CREATED

    # THEN the analysis was persisted
    analysis_id = int(response.json["id"])
    assert store.get_analysis_with_id(analysis_id)



from trailblazer.dto.failed_jobs_response import FailedJobsResponse
from trailblazer.dto.job_response import JobResponse
from trailblazer.services.slurm.dtos import SlurmJobInfo
from trailblazer.store.models import Job


def create_job_response(job: Job) -> JobResponse:
    return JobResponse(
        slurm_id=job.slurm_id, analysis_id=job.analysis_id, status=job.status, id=job.id
    )


def create_failed_jobs_response(failed_job_statistics: list[dict]) -> FailedJobsResponse:
    return FailedJobsResponse(jobs=failed_job_statistics)


def slurm_info_to_job(slurm_job_info: SlurmJobInfo) -> Job:
    return Job(
        slurm_id=slurm_job_info.slurm_id,
        name=slurm_job_info.name,
        status=slurm_job_info.status,
        elapsed=slurm_job_info.elapsed,
        started_at=slurm_job_info.started_at,
    )



from trailblazer.constants import TrailblazerStatus
from trailblazer.constants import Workflow
from trailblazer.dto import (
    AnalysesRequest,
    AnalysesResponse,
    AnalysisResponse,
    AnalysisUpdateRequest,
    CreateAnalysisRequest,
)
from trailblazer.dto.analyses_response import UpdateAnalysesResponse
from trailblazer.dto.summaries_request import SummariesRequest
from trailblazer.dto.summaries_response import SummariesResponse, Summary
from trailblazer.dto.update_analyses import UpdateAnalyses
from trailblazer.exc import MissingAnalysis
from trailblazer.services.analysis_service.utils import (
    create_analysis_response,
    create_summary,
    create_update_analyses_response,
)
from trailblazer.services.job_service.job_service import JobService
from trailblazer.store.models import Analysis, Job, User
from trailblazer.store.store import Store


class AnalysisService:
    def __init__(self, store: Store, job_service: JobService):
        self.store = store
        self.job_service = job_service

    def get_analyses(self, request: AnalysesRequest) -> AnalysesResponse:
        analyses, total_count = self.store.get_paginated_analyses(request)
        return self.create_analyses_response(analyses=analyses, total_count=total_count)

    def get_analysis(self, analysis_id: int) -> AnalysisResponse:
        if not (analysis := self.store.get_analysis_with_id(analysis_id)):
            raise MissingAnalysis(f"Analysis with id: {analysis_id} not found")
        return create_analysis_response(analysis)

    def update_analysis(
        self, analysis_id: int, update: AnalysisUpdateRequest, user: User | None = None
    ) -> AnalysisResponse:
        analysis: Analysis = self.store.update_analysis(
            analysis_id=analysis_id,
            comment=update.comment,
            is_delivered=update.is_delivered,
            status=update.status,
            is_visible=update.is_visible,
            user=user,
        )
        return create_analysis_response(analysis)

    def update_analyses(
        self, data: UpdateAnalyses, user: User | None = None
    ) -> UpdateAnalysesResponse:
        analyses: list[Analysis] = self.store.update_analyses(data=data, user=user)
        return create_update_analyses_response(analyses)

    def add_pending_analysis(self, request_data: CreateAnalysisRequest) -> AnalysisResponse:
        analysis: Analysis = self.store.add_pending_analysis(request_data)
        return create_analysis_response(analysis)

    def create_analyses_response(
        self, analyses: list[Analysis], total_count: int
    ) -> AnalysesResponse:
        response_data: list[dict] = []
        for analysis in analyses:
            analysis_data = analysis.to_dict()
            failed_job: Job = self.store.get_latest_failed_job_for_analysis(analysis.id)
            analysis_data["failed_job"] = failed_job.to_dict() if failed_job else None
            response_data.append(analysis_data)
        return AnalysesResponse(analyses=response_data, total_count=total_count)

    def update_ongoing_analyses(self) -> None:
        analyses: list[Analysis] = self.store.get_ongoing_analyses()
        for analysis in analyses:
            self.job_service.update_jobs(analysis.id)
            status: TrailblazerStatus = self.job_service.get_analysis_status(analysis.id)
            progress: float = self.job_service.get_analysis_progression(analysis.id)
            self.store.update_analysis_progress(analysis_id=analysis.id, progress=progress)
            self.store.update_analysis_status(analysis_id=analysis.id, status=status)

    def get_summaries(self, request_data: SummariesRequest) -> SummariesResponse:
        summaries: list[Summary] = []
        for order_id in request_data.order_ids:
            analyses: list[Analysis] = self.store.get_latest_analyses_for_order(order_id)
            summary: Summary = create_summary(analyses=analyses, order_id=order_id)
            summaries.append(summary)
        return SummariesResponse(summaries=summaries)



from trailblazer.apps.slurm.models import SqueueJob, SqueueResult
from trailblazer.clients.slurm_cli_client.slurm_cli_client import SlurmCLIClient
from trailblazer.clients.slurm_cli_client.utils import create_job_info_dto
from trailblazer.services.slurm.dtos import SlurmJobInfo
from trailblazer.services.slurm.slurm_service import SlurmService


class SlurmCLIService(SlurmService):
    def __init__(self, client: SlurmCLIClient):
        self.client = client

    def get_job(self, job_id: int) -> SlurmJobInfo:
        queue: SqueueResult = self.client.get_slurm_queue(str(job_id))
        job: SqueueJob = queue.jobs[0]
        return create_job_info_dto(job)

    def get_jobs(self, job_ids: list[int]) -> list[SlurmJobInfo]:
        queue: SqueueResult = self.client.get_slurm_queue(*map(str, job_ids))
        return [create_job_info_dto(job) for job in queue.jobs]



import datetime
from flask import Request

from trailblazer.dto import AnalysesRequest


def parse_analyses_request(request: Request) -> AnalysesRequest:
    """Parse a request for retrieving analyses."""
    query_params = {}
    for key in request.args.keys():
        if key_has_list_of_values(key):
            query_params[key[:-2]] = request.args.getlist(key)
        else:
            query_params[key] = request.args.get(key)
    return AnalysesRequest.model_validate(query_params)


def key_has_list_of_values(key: str) -> bool:
    return key.endswith("[]")


def stringify_timestamps(data: dict) -> dict[str, str]:
    """Convert datetime into string before dumping in order to avoid information loss"""
    for key, val in data.items():
        if isinstance(val, datetime.datetime):
            data[key] = str(val)
    return data



from trailblazer.constants import TrailblazerStatus
from trailblazer.dto.update_analyses import AnalysisUpdate, UpdateAnalyses
from trailblazer.services.analysis_service.analysis_service import AnalysisService
from trailblazer.store.models import Analysis, User
from trailblazer.store.store import Store


def test_patch_analyses_delivered(
    analysis_store: Store, analysis_id: int, analysis_service: AnalysisService
):
    analysis: Analysis = analysis_store.get_analysis_with_id(analysis_id)
    analysis_update = AnalysisUpdate(id=analysis_id, is_delivered=True)
    update_analyses = UpdateAnalyses(analyses=[analysis_update])
    user = User(id=1)

    analysis_service.update_analyses(data=update_analyses, user=user)

    assert analysis.delivered_by


def test_update_analysis_status(analysis_store: Store, analysis_service: AnalysisService):
    # GIVEN a store with ongoing analyses
    assert analysis_store.get_ongoing_analyses()
    # GIVEN that all slurm jobs are completed

    # WHEN updating the status of the analyses
    analysis_service.update_ongoing_analyses()

    # THEN the status of all analyses should be updated
    assert not analysis_store.get_ongoing_analyses()



from pydantic import BaseModel


class Summary(BaseModel):
    order_id: int
    total: int

    cancelled: int
    completed: int
    delivered: int
    failed: int
    running: int


class SummariesResponse(BaseModel):
    summaries: list[Summary]



from flask.testing import FlaskClient
from http import HTTPStatus

from trailblazer.store.models import Analysis


def test_get_existing_analysis(client: FlaskClient, analysis: Analysis):
    # GIVEN an analysis

    # WHEN requesting the analysis
    response = client.get(f"/api/v1/analyses/{analysis.id}")

    # THEN it gives a success response
    assert response.status_code == HTTPStatus.OK

    # THEN it should return the analysis
    assert response.json["id"] == analysis.id


def test_get_non_existing_analysis(client: FlaskClient, non_existing_analysis_id: str):
    # GIVEN a non existing analysis id

    # WHEN requesting the analysis
    response = client.get(f"/api/v1/analyses/{non_existing_analysis_id}")

    # THEN no analysis is found
    assert response.status_code == HTTPStatus.NOT_FOUND



from trailblazer.constants import TrailblazerStatus
from trailblazer.dto.analyses_response import UpdateAnalysesResponse
from trailblazer.dto.analysis_response import AnalysisResponse
from trailblazer.dto.summaries_response import Summary
from trailblazer.store.models import Analysis


def get_status_counts(analyses: list[Analysis]) -> dict[TrailblazerStatus, int]:
    """Returns the amount of analyses with each status."""
    delivered: int = 0
    status_counts: dict = {status: 0 for status in TrailblazerStatus}
    for analysis in analyses:
        if analysis.delivery:
            delivered += 1
        else:
            status_counts[analysis.status] += 1
    status_counts["delivered"] = delivered
    return status_counts


def create_summary(analyses: list[Analysis], order_id: int) -> Summary:
    total: int = len(analyses)
    status_counts: dict[TrailblazerStatus, int] = get_status_counts(analyses)
    return Summary(order_id=order_id, total=total, **status_counts)


def create_analysis_response(analysis: Analysis) -> AnalysisResponse:
    analysis_data: dict = analysis.to_dict()
    analysis_data["jobs"] = [job.to_dict() for job in analysis.analysis_jobs]
    analysis_data["upload_jobs"] = [job.to_dict() for job in analysis.upload_jobs]
    analysis_data["user"] = analysis.user.to_dict() if analysis.user else None
    return AnalysisResponse.model_validate(analysis_data)


def create_update_analyses_response(analyses: list[Analysis]) -> UpdateAnalysesResponse:
    response_data: list[dict] = []
    for analysis in analyses:
        analysis_data = analysis.to_dict()
        response_data.append(analysis_data)
    return UpdateAnalysesResponse(analyses=response_data)



from unittest import mock
from trailblazer.constants import SlurmJobStatus
from trailblazer.services.job_service import JobService
from trailblazer.services.slurm.dtos import SlurmJobInfo
from trailblazer.store.models import Analysis


def test_update_upload_jobs(
    job_service: JobService, upload_job_info: SlurmJobInfo, analysis: Analysis
):
    # GIVEN a store with an analysis with an ongoing upload job

    # GIVEN that slurm says the upload job is completed
    slurm_service_mock = mock.Mock()
    slurm_service_mock.get_job.return_value = upload_job_info
    job_service.slurm_service = slurm_service_mock

    # WHEN updating the upload jobs
    job_service.update_upload_jobs()

    # THEN the upload job is marked as completed
    analysis: Analysis = job_service.store.get_analysis_with_id(analysis.id)
    assert analysis.upload_jobs
    assert analysis.upload_jobs[0].status == SlurmJobStatus.COMPLETED



from datetime import datetime
from typing import Callable

from sqlalchemy import desc
from sqlalchemy.orm import Query

from trailblazer.constants import JobType, TrailblazerStatus
from trailblazer.dto.analyses_request import AnalysesRequest
from trailblazer.store.base import BaseHandler
from trailblazer.store.filters.analyses_filters import AnalysisFilter, apply_analysis_filter
from trailblazer.store.filters.job_filters import JobFilter, apply_job_filters
from trailblazer.store.filters.user_filters import UserFilter, apply_user_filter
from trailblazer.store.models import Analysis, Job, User


class ReadHandler(BaseHandler):
    """Class for reading items in the database."""

    def get_failed_jobs_stats(self, since_when: datetime = None) -> list[dict]:
        """Return the number of failed jobs per category (name)."""
        return self.get_nr_jobs_with_status_per_category(
            status=TrailblazerStatus.FAILED, since_when=since_when
        )

    def get_nr_jobs_with_status_per_category(
        self, status: str, since_when: datetime = None
    ) -> list[dict[str, str | int]]:
        """Return the number of jobs with status per category (name)."""
        filter_map: dict[Callable, str | bool | None] = {
            JobFilter.BY_STATUS: status,
            JobFilter.BY_SINCE_WHEN: since_when,
        }
        filter_functions: list[Callable] = [
            function for function, supplied_arg in filter_map.items() if supplied_arg
        ]
        categories = apply_job_filters(
            filters=filter_functions,
            jobs=self.get_job_query_with_name_and_count_labels(),
            since_when=since_when,
            status=status,
        )
        categories = categories.group_by(Job.name).all()
        return [{"name": category.name, "count": category.count} for category in categories]

    def get_analyses_by_status_started_at_and_comment(
        self,
        before: datetime | None = None,
        comment: str | None = None,
        status: str | None = None,
    ) -> list[Analysis] | None:
        """Return analyses meeting supplied arguments."""
        if not before and not comment and not status:
            return
        filter_map: dict[Callable, str | datetime | TrailblazerStatus | None] = {
            AnalysisFilter.BY_COMMENT: comment,
            AnalysisFilter.BY_BEFORE_STARTED_AT: before,
            AnalysisFilter.BY_STATUS: status,
        }
        filter_functions: list[Callable] = [
            function for function, supplied_arg in filter_map.items() if supplied_arg
        ]
        analyses: Query = apply_analysis_filter(
            filter_functions=filter_functions,
            analyses=self.get_query(table=Analysis),
            comment=comment,
            started_at=before,
            status=status,
        )
        return analyses.order_by(desc(Analysis.started_at)).all()

    def get_analysis(self, case_id: str, started_at: datetime, status: str) -> Analysis | None:
        """Return the latest analysis for supplied parameters."""
        return apply_analysis_filter(
            analyses=self.get_query(table=Analysis),
            filter_functions=[
                AnalysisFilter.BY_CASE_ID,
                AnalysisFilter.BY_STARTED_AT,
                AnalysisFilter.BY_STATUS,
            ],
            case_id=case_id,
            started_at=started_at,
            status=status,
        ).first()

    def get_latest_analysis_for_case(self, case_id: str) -> Analysis | None:
        """Return the latest analysis for a case."""
        return (
            apply_analysis_filter(
                analyses=self.get_query(table=Analysis),
                filter_functions=[AnalysisFilter.BY_CASE_ID],
                case_id=case_id,
            )
            .order_by(desc(Analysis.started_at))
            .first()
        )

    def get_analyses_for_case(self, case_id: str) -> list[Analysis] | None:
        """Return all analyses for a case."""
        return apply_analysis_filter(
            analyses=self.get_query(table=Analysis),
            filter_functions=[AnalysisFilter.BY_CASE_ID],
            case_id=case_id,
        ).all()

    def get_analyses_with_statuses(self, statuses: list[str]) -> list[Analysis] | None:
        """Get analyses by statuses."""
        return apply_analysis_filter(
            analyses=self.get_query(table=Analysis),
            filter_functions=[AnalysisFilter.BY_STATUSES],
            statuses=statuses,
        ).all()

    def get_analysis_with_id(self, analysis_id: int) -> Analysis | None:
        """Get a single analysis by id."""
        return apply_analysis_filter(
            analyses=self.get_query(table=Analysis),
            filter_functions=[AnalysisFilter.BY_ENTRY_ID],
            analysis_id=analysis_id,
        ).first()

    def get_user(
        self,
        email: str = None,
        exclude_archived: bool = True,
    ) -> User:
        """Return user from the database."""
        filter_map: dict[Callable, str | bool | None] = {
            UserFilter.BY_CONTAINS_EMAIL: email,
            UserFilter.BY_IS_NOT_ARCHIVED: exclude_archived,
        }
        filter_functions: list[Callable] = [
            function for function, supplied_arg in filter_map.items() if supplied_arg
        ]
        return apply_user_filter(
            filter_functions=filter_functions,
            users=self.get_query(table=User),
            email=email,
        ).first()

    def get_user_by_id(self, user_id: int) -> User | None:
        return apply_user_filter(
            filter_functions=[UserFilter.BY_ID],
            users=self.get_query(table=User),
            id=user_id,
        ).first()

    def get_users(
        self,
        name: str = None,
        email: str = None,
        exclude_archived: bool = True,
    ) -> list[User]:
        """Return users from the database."""
        filter_map: dict[Callable, str | bool | None] = {
            UserFilter.BY_CONTAINS_EMAIL: email,
            UserFilter.BY_CONTAINS_NAME: name,
            UserFilter.BY_IS_NOT_ARCHIVED: exclude_archived,
        }
        filter_functions: list[Callable] = [
            function for function, supplied_arg in filter_map.items() if supplied_arg
        ]
        return apply_user_filter(
            filter_functions=filter_functions,
            users=self.get_query(table=User),
            email=email,
            name=name,
        ).all()

    def get_latest_failed_job_for_analysis(self, analysis_id: str) -> Job | None:
        filters: list[Callable] = [
            JobFilter.BY_ANALYSIS_ID,
            JobFilter.BY_STATUS,
            JobFilter.SORT_BY_STARTED_AT,
        ]
        return apply_job_filters(
            filters=filters,
            jobs=self.get_query(Job),
            analysis_id=analysis_id,
            status=TrailblazerStatus.FAILED,
        ).first()

    def get_ongoing_upload_jobs(self) -> list[Job]:
        ongoing_statuses: list[str] = list(TrailblazerStatus.ongoing_statuses())
        return apply_job_filters(
            filters=[JobFilter.BY_TYPE, JobFilter.BY_STATUSES],
            jobs=self.get_query(Job),
            job_type=JobType.UPLOAD,
            statuses=ongoing_statuses,
        ).all()

    def get_job_by_id(self, job_id: int) -> Job | None:
        return apply_job_filters(
            filters=[JobFilter.BY_ID],
            jobs=self.get_query(Job),
            job_id=job_id,
        ).first()

    def get_latest_analyses_for_order(self, order_id: int) -> list[Analysis]:
        """Returns the latest analysis per case in the given order."""
        return apply_analysis_filter(
            analyses=self.get_query(Analysis),
            filter_functions=[
                AnalysisFilter.BY_ORDER_ID,
                AnalysisFilter.BY_LATEST_PER_CASE,
                AnalysisFilter.EXCLUDE_RSYNC,
            ],
            order_id=order_id,
        ).all()

    def get_paginated_analyses(self, request: AnalysesRequest) -> tuple[list[Analysis], int]:
        analyses: Query = self._filter_analyses(request)
        total_count: int = analyses.count()
        page: Query = self._paginate_analyses(analyses=analyses, request=request)
        return page.all(), total_count

    def _filter_analyses(self, request: AnalysesRequest) -> Query:
        filters: list[AnalysisFilter] = [
            AnalysisFilter.BY_WORKFLOW,
            AnalysisFilter.BY_HAS_COMMENT,
            AnalysisFilter.BY_ORDER_ID,
            AnalysisFilter.BY_PRIORITIES,
            AnalysisFilter.BY_STATUSES,
            AnalysisFilter.BY_TYPES,
            AnalysisFilter.BY_CASE_ID,
            AnalysisFilter.BY_SEARCH_TERM,
            AnalysisFilter.BY_IS_VISIBLE,
            AnalysisFilter.BY_DELIVERED,
            AnalysisFilter.SORTING,
        ]
        return apply_analysis_filter(
            filter_functions=filters,
            analyses=self.get_query(Analysis),
            has_comment=request.has_comment,
            order_id=request.order_id,
            priorities=request.priority,
            statuses=request.status,
            types=request.type,
            case_id=request.case_id,
            workflow=request.workflow,
            search_term=request.search,
            show_hidden=bool(request.search),
            sort_field=request.sort_field,
            sort_order=request.sort_order,
            delivered=request.delivered,
        )

    def _paginate_analyses(self, analyses: Query, request: AnalysesRequest) -> Query:
        return apply_analysis_filter(
            filter_functions=[AnalysisFilter.PAGINATION],
            analyses=analyses,
            page=request.page,
            page_size=request.page_size,
        )



from trailblazer.clients.slurm_api_client.dto.cancel_job_response import (
    SlurmCancelJobResponse,
)  # noqa: F401
from trailblazer.clients.slurm_api_client.dto.job_response import (
    SlurmJobResponse,
)  # noqa: F401
from trailblazer.clients.slurm_api_client.dto.jobs_response import (
    SlurmJobsResponse,
)  # noqa: F401



import datetime

from sqlalchemy.orm import Query

from tests.mocks.store_mock import MockStore
from trailblazer.store.filters.analyses_filters import (
    filter_analyses_by_before_started_at,
    filter_analyses_by_case_id,
    filter_analyses_by_comment,
    filter_analyses_by_entry_id,
    filter_analyses_by_is_visible,
    filter_analyses_by_search_term,
    filter_analyses_by_started_at,
    filter_analyses_by_status,
    filter_analyses_by_statuses,
)
from trailblazer.store.models import Analysis


def test_filter_analyses_by_id(analysis_store: MockStore):
    """Test return analysis by id when existing."""
    # GIVEN a store containing analyses
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # WHEN retrieving an analysis by id
    analysis: Query = filter_analyses_by_entry_id(
        analyses=analysis_store.get_query(table=Analysis), analysis_id=existing_analysis.id
    )

    # THEN that the analysis is a query
    assert isinstance(analysis, Query)

    # THEN the analysis should match the original
    assert existing_analysis == analysis.first()


def test_filter_analyses_by_case_id(analysis_store: MockStore):
    """Test return analysis by case id when existing."""
    # GIVEN a store containing analyses
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # WHEN retrieving an analysis by case id
    analysis: Query = filter_analyses_by_case_id(
        analyses=analysis_store.get_query(table=Analysis), case_id=existing_analysis.case_id
    )

    # THEN that the analysis is a query
    assert isinstance(analysis, Query)

    # THEN the analysis should match the original
    assert existing_analysis == analysis.first()


def test_filter_analyses_by_comment(analysis_store: MockStore, analysis_comment: str):
    """Test return analysis by comment when containing string."""
    # GIVEN a store containing analyses
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # GIVEN a comment
    existing_analysis.comment = analysis_comment

    # WHEN retrieving an analysis by comment
    analysis: Query = filter_analyses_by_comment(
        analyses=analysis_store.get_query(table=Analysis), comment=existing_analysis.comment
    )

    # THEN that the analysis is a query
    assert isinstance(analysis, Query)

    # THEN the analysis should match the original
    assert existing_analysis == analysis.first()


def test_filter_analyses_by_comment_when_not_matching(analysis_store: MockStore):
    """Test return analysis by comment when not containing string."""
    # GIVEN a store containing analyses

    # GIVEN a comment that does not exist in the database

    # WHEN retrieving an analysis by comment
    analysis: Query = filter_analyses_by_comment(
        analyses=analysis_store.get_query(table=Analysis), comment="Does not exist."
    )

    # THEN no analysis is returned
    assert not analysis.first()


def test_filter_analyses_by_is_visible(analysis_store: MockStore):
    """Test return analysis when is visible is true."""
    # GIVEN a store containing analyses
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # GIVEN a visible analysis
    existing_analysis.is_visible = True

    # GIVEN an analysis query
    analyses: Query = analysis_store.get_query(table=Analysis)

    # WHEN retrieving analyses by is visible
    analyses: Query = filter_analyses_by_is_visible(analyses=analyses, show_hidden=True)

    # THEN the analyses is a query
    assert isinstance(analyses, Query)

    # THEN the analysis should match the original
    assert existing_analysis == analyses.first()


def test_filter_analyses_by_is_visible_when_false(analysis_store: MockStore):
    """Test return analysis when is visible is false."""
    # GIVEN a store containing analyses
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # GIVEN a hidden analysis
    existing_analysis.is_visible = False

    # GIVEN an analysis query
    analyses: Query = analysis_store.get_query(table=Analysis)

    # WHEN retrieving analyses by is visible
    analyses: Query = filter_analyses_by_is_visible(analyses=analyses, show_hidden=False)

    # THEN the analyses is a query
    assert isinstance(analyses, Query)

    # THEN the existing analysis should not be returned
    for analysis in analyses:
        assert existing_analysis != analysis


def test_filter_analyses_by_search_term_in_comment(
    analysis_store: MockStore, analysis_comment: str
):
    """Test return analysis when search term matches comment."""
    # GIVEN a store containing analyses
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # GIVEN a a comment
    existing_analysis.comment = analysis_comment

    # WHEN retrieving analyses by search term
    analyses: Query = filter_analyses_by_search_term(
        analyses=analysis_store.get_query(table=Analysis), search_term=analysis_comment
    )

    # THEN the analysis is a query
    assert isinstance(analyses, Query)

    # THEN the analysis should match the original
    assert existing_analysis == analyses.first()


def test_filter_analyses_by_search_term_in_workflow(analysis_store: MockStore):
    """Test return analysis when search term matches workflow."""
    # GIVEN a store containing analyses
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # WHEN retrieving analyses by search term
    analyses: Query = filter_analyses_by_search_term(
        analyses=analysis_store.get_query(table=Analysis),
        search_term=existing_analysis.workflow,
    )

    # THEN the analysis is a query
    assert isinstance(analyses, Query)

    # THEN the analysis should match the original
    assert existing_analysis == analyses.first()


def test_filter_analyses_by_search_term_in_case(analysis_store: MockStore):
    """Test return analysis when search term matches case."""
    # GIVEN a store containing analyses
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # WHEN retrieving analyses by search term
    analyses: Query = filter_analyses_by_search_term(
        analyses=analysis_store.get_query(table=Analysis), search_term=existing_analysis.case_id
    )

    # THEN the analysis is a query
    assert isinstance(analyses, Query)

    # THEN the analysis should match the original
    assert existing_analysis == analyses.first()


def test_filter_analyses_by_search_term_in_status(analysis_store: MockStore):
    """Test return analysis when search term matches status."""
    # GIVEN a store containing analyses
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # WHEN retrieving analyses by search term
    analyses: Query = filter_analyses_by_search_term(
        analyses=analysis_store.get_query(table=Analysis), search_term=existing_analysis.status
    )

    # THEN the analyses is a query
    assert isinstance(analyses, Query)

    # THEN the analysis should match the original
    assert existing_analysis == analyses.first()


def test_filter_analyses_by_search_term_when_not_matching(analysis_store: MockStore):
    """Test return analysis when search term does not match."""
    # GIVEN a store containing analyses
    analysis_store.get_query(table=Analysis).first()

    # WHEN retrieving analyses by search term
    analyses: Query = filter_analyses_by_search_term(
        analyses=analysis_store.get_query(table=Analysis),
        search_term="does not exist",
    )
    # THEN the analyses is a query
    assert isinstance(analyses, Query)

    # THEN the no analysis is returned
    assert not analyses.first()


def test_filter_analyses_by_started_at(analysis_store: MockStore):
    """Test return analysis by started at when existing."""
    # GIVEN a store containing analyses
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # WHEN retrieving an analysis by when it was started
    analysis: Query = filter_analyses_by_started_at(
        analyses=analysis_store.get_query(table=Analysis), started_at=existing_analysis.started_at
    )

    # THEN that the analysis is a query
    assert isinstance(analysis, Query)

    # THEN the analysis should match the original
    assert existing_analysis == analysis.first()


def test_filter_analyses_by_before_started_at(analysis_store: MockStore, timestamp_now: datetime):
    """Test return analysis with a started at before supplied date when existing."""
    # GIVEN a store containing analyses

    # WHEN retrieving an analysis before supplied date
    analysis: Query = filter_analyses_by_before_started_at(
        analyses=analysis_store.get_query(table=Analysis), started_at=timestamp_now
    )

    # THEN that the analysis is a query
    assert isinstance(analysis, Query)

    # THEN the analysis should return an analysis before than supplied date
    assert analysis.first().started_at < timestamp_now


def test_filter_analyses_by_before_started_at_when_after_supplied_date(
    analysis_store: MockStore, timestamp_old: datetime
):
    """Test return analysis with a started at after supplied date when existing."""
    # GIVEN a store containing analyses

    # WHEN retrieving an analysis by when it was started
    analysis: Query = filter_analyses_by_before_started_at(
        analyses=analysis_store.get_query(table=Analysis), started_at=timestamp_old
    )

    # THEN that the analysis is a query
    assert isinstance(analysis, Query)

    # THEN no analysis is returned
    assert not analysis.first()


def test_filter_analyses_by_status(analysis_store: MockStore):
    """Test return analysis by status when existing."""
    # GIVEN a store containing analyses
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # WHEN retrieving an analysis by status
    analysis: Query = filter_analyses_by_status(
        analyses=analysis_store.get_query(table=Analysis), status=existing_analysis.status
    )

    # THEN that the analysis is a query
    assert isinstance(analysis, Query)

    # THEN the analysis should match the original
    assert existing_analysis == analysis.first()


def test_filter_analyses_by_statuses(analysis_store: MockStore):
    """Test return analyses by statuses when existing."""
    # GIVEN a store containing analyses
    existing_analysis: Analysis = analysis_store.get_query(table=Analysis).first()

    # WHEN retrieving an analysis by status
    analysis: Query = filter_analyses_by_statuses(
        analyses=analysis_store.get_query(table=Analysis), statuses=[existing_analysis.status]
    )

    # THEN that the analysis is a query
    assert isinstance(analysis, Query)

    # THEN the analysis should match the original
    assert existing_analysis == analysis.first()



from pydantic import BaseModel

from trailblazer.clients.slurm_api_client.dto.common import SlurmError, SlurmJobInfo, SlurmWarning


class SlurmJobsResponse(BaseModel):
    jobs: list[SlurmJobInfo]
    errors: list[SlurmError] | None = None
    warnings: list[SlurmWarning] | None = None



from pathlib import Path
from typing import Any

import json
from trailblazer.constants import FileExtension
from trailblazer.io.validate_path import validate_file_suffix


def read_json(file_path: Path) -> Any:
    """Read content in a JSON file."""
    validate_file_suffix(path_to_validate=file_path, target_suffix=FileExtension.JSON)
    with open(file_path, "r") as file:
        return json.load(file)
